/*
Tangle Tail Grapple
	by Lach and SMS Alfredo
	
(the typo in the lump name was accidental but I'm keeping it)
*/

local SKIN = "tangle"				// player skin to run script for
local TAILSKIN = "tangle"		// skin to apply to tail

local THROW_DISTANCE = 512*FRACUNIT	// maximum distance the tail can be thrown before giving up and returning
local STRETCH_CAP = 2048*FRACUNIT	// (sort of) maximum stretch distance for the tail
local STRETCH_THRESHOLD = 96*FRACUNIT	// distance from tail's tip where pulling starts to take place
local SEG_TIME = TICRATE*5 // how long each tail segment object will last for when not being used

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			freeslot(item)
		end
	end
end

//Command to enable diving on water
local tanglepickup = CV_RegisterVar({
	name = "tanglepickup",
	defaultvalue = "On",
	flags = CV_NETVAR|CV_SHOWMODIF,
	PossibleValue = CV_OnOff
})

//tangleswap command
local function tangleswap(player, arg)
	if not (player.mo and player.mo.valid) then CONS_Printf(player, "This can only be used in a level!") return end
	if arg
		if arg == "0" or arg == "off" or arg == "default"
			player.tangleswap = 0
			if io and player == consoleplayer
				local file = io.openlocal("client/tangle.dat", "w+")
				file:write(0)
				file:close()
			end
		elseif arg == "1" or arg == "on"
			player.tangleswap = 1
			if io and player == consoleplayer
				local file = io.openlocal("client/tangle.dat", "w+")
				file:write(1)
				file:close()
			end
		elseif player.tangleswap
			CONS_Printf(player, "tangleswap - Swaps Tangle's tail abilities on the ground. Currently on.")
		else
			CONS_Printf(player, "tangleswap - Swaps Tangle's tail abilities on the ground. Currently off.")
		end
	elseif player.tangleswap
		CONS_Printf(player, "tangleswap - Swaps Tangle's tail abilities on the ground. Currently on.")
	else
		CONS_Printf(player, "tangleswap - Swaps Tangle's tail abilities on the ground. Currently off.")
	end
end
COM_AddCommand("tangleswap", tangleswap)
COM_AddCommand("tangleswap2", tangleswap, COM_SPLITSCREEN)

addHook("PlayerSpawn", function(player)
	//I/O maybe
	if io and player.tangleswap == nil and player == consoleplayer
		local file = io.openlocal("client/tangle.dat")
		if file
			local num = file:read("*n")
			if num == 0 or num == 1
				COM_BufInsertText(consoleplayer, "tangleswap "..num)
			end
			file:close()
		end
	end
end)

SafeFreeslot("S_TANGRAPPLE_HAND", "MT_TANGRAPPLE_HAND",
"S_TANGRAPPLE_TAIL", "MT_TANGRAPPLE_TAIL",
"S_TANGRAPPLE_INVISIBLE", "S_TANGRAPPLE_SEGMENT", "S_TANGRAPPLE_SUPERSEGMENT", "MT_TANGRAPPLE_SEGMENT",
"sfx_tngjm2", "sfx_tngjm3")

sfxinfo[sfx_tngjm2].caption = sfxinfo[sfx_jump].caption
sfxinfo[sfx_tngjm3].caption = sfxinfo[sfx_jump].caption

states[S_TANGRAPPLE_HAND] = {
	sprite = SPR_PLAY,
	frame = SPR2_TAL0,
	tics = -1
}

mobjinfo[MT_TANGRAPPLE_HAND] = {
	name = "elastic tail",
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	seestate = S_TANGRAPPLE_HAND,
	height = 16*FRACUNIT,
	radius = 8*FRACUNIT,
	flags = MF_NOGRAVITY,
	speed = 32*FRACUNIT
}

states[S_TANGRAPPLE_TAIL] = {
	sprite = SPR_PLAY,
	frame = SPR2_TAL2,
	tics = -1
}

mobjinfo[MT_TANGRAPPLE_TAIL] = {
	name = "elastic tail",
	spawnstate = S_INVISIBLE,
	spawnhealth = 1000,
	seestate = S_TANGRAPPLE_TAIL,
	height = 16*FRACUNIT,
	radius = 8*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY,
}

states[S_TANGRAPPLE_INVISIBLE] = {
	tics = SEG_TIME,
	
	var1 = MF_NOSECTOR | MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY|MF_NOCLIP|MF_NOCLIPTHING,
	var2 = 0,
	action = A_SetObjectFlags,
}

states[S_TANGRAPPLE_SEGMENT] = {
	sprite = SPR_PLAY,
	frame = SPR2_TAL1,
	tics = 1,
	nextstate = S_TANGRAPPLE_INVISIBLE,
	
	var1 = MF_NOSECTOR,
	var2 = 1,
	action = A_SetObjectFlags,
}

states[S_TANGRAPPLE_SUPERSEGMENT] = {
	sprite = SPR_PLAY,
	frame = SPR2_TAL1|FF_FULLBRIGHT,
	tics = 1,
	nextstate = S_TANGRAPPLE_INVISIBLE,
	
	var1 = MF_NOSECTOR,
	var2 = 1,
	action = A_SetObjectFlags,
}

mobjinfo[MT_TANGRAPPLE_SEGMENT] = {
	name = "elastic tail",
	spawnstate = S_TANGRAPPLE_INVISIBLE,
	seestate = S_TANGRAPPLE_SEGMENT,
	meleestate = S_TANGRAPPLE_SUPERSEGMENT,
	height = FRACUNIT,
	radius = FRACUNIT,
	damage = 1,
	flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_SCENERY|MF_NOCLIP|MF_NOCLIPTHING
}

local P_TeleportMove = P_MoveOrigin or P_TeleportMove -- minimal effort warning suppressor

---------------

local TF_COUNTJUMP = 1 << 0 // allow the jump button to sustain a grapple
local TF_FLOOR = 1 << 1 // the grappled surface is the floor
local TF_AIRBORNE = 1 << 2 // allow grappling in mid-air, even when not jumping

// because fuuuuuck typing this over and over
local function valid(mo)
	return mo and mo.valid
end

// squares two fixed point numbers
local function FixedSquare(n)
	return FixedMul(n, n)
end

// returns whether grappling is permitted in the player's current circumstances
local function PlayerCanGrapple(player)
	return player.mo and player.mo.valid and player.mo.health
	and not P_PlayerInPain(player)
	and not player.playerstate
	and not (player.pflags & PF_SLIDING)
	and not player.exiting
	and not player.powers[pw_carry]
	and not player.spectator
	and not player.playerstate
	and player.powers[pw_super] != 1
	and not player.iseggrobo
	and not player.isjettysyn
	and not player.guard
	and not player.actionstate
	and not player.powers[pw_nocontrol]
	and (player.exhaustmeter == nil or player.exhaustmeter > 0)
	and not (player.isxmomentum and player.mo.state == S_PLAY_FACEPLANT)
end

// returns whether we're holding enough into a direction or not
local function PlayerFullInput(player)
	return FixedHypot(player.cmd.forwardmove*FRACUNIT, player.cmd.sidemove*FRACUNIT) > 35*FRACUNIT
end

// returns true if the object is on a bustable FOF, and breaks it
local function BustGround(mo)
	local sector = mo.subsector.sector
	local flip = mo.eflags & MFE_VERTICALFLIP
	
	for fof in sector.ffloors()
		if not fof.valid
		or (fof.flags & (FF_EXISTS|FF_BUSTUP)) ~= (FF_EXISTS|FF_BUSTUP)
			continue
		end
		
		local slope = flip and fof.b_slope or not flip and fof.t_slope or nil
		
		if slope
			if slope == mo.standingslope
				EV_CrumbleChain(sector, fof)
				return true
			end
		else
			if flip
				if fof.bottomheight == mo.ceilingz
					EV_CrumbleChain(sector, fof)
					return true
				end
			else
				if fof.topheight == mo.floorz
					EV_CrumbleChain(sector, fof)
					return true
				end
			end
		end
	end
	
	return false
end

//Why do we need a separate thing for busting walls?
local fofdetect = function(mobj, sectorthing, flag)
	if not valid(mobj) or not valid(sectorthing) return end
	for fof in sectorthing.ffloors()
		if fof.valid and (fof.flags & flag) and (fof.flags & FF_EXISTS)
			local foffloor = fof.topheight
			local fofceiling = fof.bottomheight
			if fof.b_slope and fof.b_slope.valid
				fofceiling = P_GetZAt(fof.b_slope, mobj.x, mobj.y)
			end
			if fof.t_slope and fof.t_slope.valid
				foffloor = P_GetZAt(fof.t_slope, mobj.x, mobj.y)
			end
			if foffloor >= mobj.z and fofceiling <= mobj.z+mobj.height
				if flag == FF_BUSTUP
					EV_CrumbleChain(nil, fof)
					if fof.master.flags & ML_EFFECT5
						P_LinedefExecute(FixedHypot(fof.master.dx, fof.master.dy)>>FRACBITS, mobj, sectorthing)
					end
				end
				return true
			end
		end
	end
end

//P_LookForEnemies still kinda sucks
local EnemyLookDeluxe = function(player, distance, span)
	if not (player.mo and player.mo.valid) return end
	local maxdist = FixedMul(distance, player.mo.scale)
	local thokmobj = P_SpawnMobjFromMobj(player.mo, 0,0,0, MT_THOK)
	thokmobj.radius = maxdist
	local closestmo = nil
	local closestdist = 0

	searchBlockmap("objects", function(refmobj, mo)
		//Prevent targeting Battle's target dummy since we can target players anyways
		if (pcall(function() local v = _G["MT_TARGETDUMMY"] end, "MT_TARGETDUMMY")
		and mo.type == _G["MT_TARGETDUMMY"]) then return end
		
		local trash = R_PointToAngle2(player.mo.x, player.mo.y, mo.x, mo.y)
		
		if mo == player.mouthmobj or (mo.flags & MF_NOCLIPTHING)
		or mo.mouthed or mo.health <= 0
		or mo == player.mo or mo.flags2 & MF2_FRET or not P_CheckSight(player.mo, mo)
		or ((twodlevel or player.mo.flags2 & MF2_TWOD) and abs(player.mo.y-mo.y) > player.mo.radius)
		or ((trash - player.mo.angle > span) or (trash - player.mo.angle < -1*span))
		return end

		if mo.player
			if not tanglepickup.value
				return
			elseif mo.player.spectator
				return
			elseif G_GametypeHasTeams() and player.ctfteam == mo.player.ctfteam
				return
			end
		elseif mo.type == MT_SPIKE or mo.type == MT_WALLSPIKE
			if valid(closestmo)
			and closestmo.type != MT_SPIKE and closestmo.type != MT_WALLSPIKE
				return
			end
		elseif mo.flags & (MF_SPRING|MF_PUSHABLE)
			//lol nothing
		elseif (mo.type == MT_RING_REDBOX and player.ctfteam != 1)
		or (mo.type == MT_RING_BLUEBOX and player.ctfteam != 2)
			return // no targeting monitors you can't destroy
		elseif (!((mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR) and (mo.flags & MF_SHOOTABLE))) == !(mo.flags2 & MF2_INVERTAIMABLE)) // allows if it has the flags desired XOR it has the invert aimable flag
			return // not a valid target
		end
		
		local zdist = ((player.mo.z + player.mo.height/2) - (mo.z + mo.height/2))*2
		local dist = FixedHypot(player.mo.x-mo.x, player.mo.y-mo.y)
		if ((R_PointToAngle2(0, 0, dist, zdist) + span) > span*2) and ((R_PointToAngle2(0, 0, dist, -1*zdist) + span) > span*2)
			return // Don't home outside of desired angle!
		end

		dist = FixedHypot(dist, zdist)
		if (dist > maxdist) or (closestmo and dist > closestdist)
			return // out of range
		end

		closestmo = mo
		closestdist = dist
	end, thokmobj)
	
	P_RemoveMobj(thokmobj)
	return closestmo
end

// spawns a fuckton of thok mobjs to display the tail
local STARTSCALE = FRACUNIT/5
local RADIUS = 16*FRACUNIT
local SEGCAP = 100 // we *shouldn't* have an issue with infinite loops, but better safe than sorry
local function RenderTail(player, hand, aim)
	local mo = player.mo
	
	// vars
	local istail = hand.type == MT_TANGRAPPLE_TAIL
	local angle = player.drawangle + ANGLE_180
	local aiming = aim or 0
	local segs = {}
	local numsegs = 0
	local scale = FixedMul(STARTSCALE, hand.scale)
	local x = mo.x + FixedMul(2*mo.scale, cos(angle))
	local y = mo.y + FixedMul(2*mo.scale, sin(angle))
	local z = mo.z
	if mo.eflags & MFE_VERTICALFLIP
		if istail
			z = $ + 3*mo.height/5
		else
			z = $ + 5*mo.height/9
		end
	else
		z = $ + 2*mo.height/5
	end
	if mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_EDGE
	or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_LAND
	or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_PAIN
	or mo.sprite2&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_RUN_|FF_SPR2SUPER
	or (player.isxmomentum and mo.state == S_PLAY_FACEPLANT)
		z = $-mo.height/8*P_MobjFlip(mo)
	elseif mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_WALK
		if mo.frame&FF_FRAMEMASK == 0
		or mo.frame&FF_FRAMEMASK == 4
			z = $-4*mo.scale*P_MobjFlip(mo)
		elseif mo.frame&FF_FRAMEMASK == 2
		or mo.frame&FF_FRAMEMASK == 6
			z = $-2*mo.scale*P_MobjFlip(mo)
		else
			z = $-3*mo.scale*P_MobjFlip(mo)
		end
	elseif mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_RUN
		z = $-(2+(mo.frame&FF_FRAMEMASK)%2)*mo.scale*P_MobjFlip(mo)
	end
	
	// constants
	local handz = hand.z + (hand.height >> 1) - (mo.eflags & MFE_VERTICALFLIP and not istail and hand.height*9/8 or 0)
	local hdistance = FixedHypot(x - hand.x, y - hand.y)
	local distance = FixedHypot(hdistance, z - handz)
	local scalespeed = FixedDiv(FixedMul(RADIUS, mo.scale), distance)
	local radius = FixedMul(RADIUS, hand.scale)
	
	if istail
		radius = $*11/8
	end
	
	// spawn stuff from the player's butt until we are close enough to the hand
	while numsegs < SEGCAP and FixedHypot(FixedHypot(x - hand.x, y - hand.y), z - handz) > radius
		local dist = FixedMul(RADIUS, scale)
		local hdist = P_ReturnThrustX(mo, aiming, dist)
		local zdist = P_ReturnThrustY(mo, aiming, dist)
		
		local hdistance = FixedHypot(x - hand.x, y - hand.y)
		local distance = FixedHypot(hdistance, z - handz)
		local goalangle = R_PointToAngle2(x, y, hand.x, hand.y)
		local goalaiming = R_PointToAngle2(0, z, hdistance, handz)
		local anglespeed = FixedDiv(scale, hand.scale)/(distance < (radius << 1) and 1 or 2)
		
		x = $ + P_ReturnThrustX(mo, angle, hdist)
		y = $ + P_ReturnThrustY(mo, angle, hdist)
		z = $ + zdist
		
		player.tanglesegs = $ or {}
		local seg = player.tanglesegs[numsegs + 1]
		if not valid(seg)
			seg = P_SpawnMobj(hand.x, hand.y, handz, MT_TANGRAPPLE_SEGMENT)
			seg.skin = TAILSKIN
			player.tanglesegs[numsegs + 1] = seg
		end
		
		if seg.tics ~= SEG_TIME and P_SetOrigin
			P_SetOrigin(seg, hand.x, hand.y, handz)
		end
		P_TeleportMove(seg, x, y, z)
		
		seg.color = mo.color
		if mo.frame & FF_FULLBRIGHT
			seg.state = seg.info.meleestate
		else
			seg.state = seg.info.seestate
		end
		
		seg.scale = scale
		seg.flags2 = ($&~MF2_OBJECTFLIP&~MF2_DONTDRAW)|(mo.flags2&(MF2_OBJECTFLIP|MF2_DONTDRAW))
		seg.eflags = ($&~MFE_VERTICALFLIP)|(mo.eflags&MFE_VERTICALFLIP)
		
		seg.target = mo
		seg.tracer = hand
		seg.pain_tics = TICRATE
		
		//Damage and scale
		if not istail
			seg.radius = scale << 4
			seg.height = scale << 4
			seg.flags = $ | MF_MISSILE &~ (MF_NOCLIP | MF_NOCLIPTHING)
		elseif not numsegs
			seg.scale = $ << 1
		end
		
		angle = $ + FixedMul(goalangle - $, anglespeed)
		aiming = $ + FixedMul(goalaiming - $, scale)
		scale = $ + FixedMul(hand.scale - $, scalespeed)
		table.insert(segs, seg)
		numsegs = $ + 1
	end
	
	for i, seg in ipairs(segs)
		//Set the colors!
		if numsegs > 4
			for j = 1, 6*i/numsegs
				seg.state = $
			end
			if istail
				seg.state = $
			end
		elseif i & 1 == 0
			seg.state = $
		end
		
		//Set the midpoint
		if i == numsegs >> 1
			player.midtanglepoint = {x = seg.x, y = seg.y, z = seg.z}
		end
	end
end

// initiate a grapple
local function StartGrapple(player, flag)
	local mo = player.mo
	local grounded = P_IsObjectOnGround(mo)
	local tangle = player.tangrapple
	
	if player.pflags & (flag|PF_JUMPSTASIS)
	or not tangle
	or not mo
	or valid(tangle.hand)
	or not (grounded or player.pflags & PF_JUMPED or tangle.flags & TF_AIRBORNE)
	or not PlayerCanGrapple(player)
	or (player.charability != CA_NONE and (flag == PF_JUMPDOWN or player.charability2 != CA2_NONE) and not grounded)
	or (player.charability2 != CA2_NONE and grounded)
	or (not (player.charflags&SF_MULTIABILITY) and not grounded and player.pflags&PF_THOKKED)
	or ((player.gotflag or player.gotcrystal) and not grounded)
	//or (player.powers[pw_shield] & SH_NOSTACK &~ SH_PITY &~ SH_PINK and not grounded
	//and not (tangle.flags & TF_AIRBORNE or (flag == PF_JUMPDOWN and not (player.pflags&PF_THOKKED))))
		return
	end
	
	mo.tailjectile = true // Bloops' variable for hiding the tail
	tangle.flags = $ & ~(TF_FLOOR|TF_COUNTJUMP)
	if flag == PF_JUMPDOWN
		tangle.flags = $ | TF_COUNTJUMP
	end
	
	local flip = P_MobjFlip(mo)
	local input = PlayerFullInput(player)
	local hand = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_TANGRAPPLE_HAND)
	local speed = FixedMul(hand.info.speed*(player.powers[pw_super] and 4 or 2), mo.scale)
	
	hand.skin = TAILSKIN
	hand.state = hand.info.seestate
	hand.tracer = mo
	hand.angle = mo.angle
	
	local lockon = EnemyLookDeluxe(player, THROW_DISTANCE*5/4, ANGLE_22h)
	if not input and lockon and lockon.valid
		P_HomingAttack(hand, lockon)
		hand.momx = hand.momx*2
		hand.momy = hand.momy*2
		hand.momz = hand.momz*2
		P_SetObjectMomZ(hand, FRACUNIT, true)
		hand.z = $+P_MobjFlip(mo)
		player.mo.angle = R_PointToAngle2(mo.x, mo.y, lockon.x, lockon.y)
		player.drawangle = player.mo.angle
		hand.flags = $|MF_NOCLIPHEIGHT
		hand.flags2 = $|MF2_AMBUSH
	elseif (not input and (not player.tangleswap or not grounded))
	or (input and not (not player.tangleswap or not grounded))
		local vspeed = FixedMul(speed, sin(ANG30))
		local hspeed = FixedMul(speed, cos(ANG30))
		if grounded
			P_SetObjectMomZ(hand, FRACUNIT, true)
			hand.flags = $|MF_NOCLIPHEIGHT
			hand.momx = mo.momx
			hand.momy = mo.momy
			P_Thrust(hand, hand.angle, hspeed)
		else
			hand.momz = -flip*vspeed + mo.momz
			P_Thrust(hand, hand.angle, hspeed/48)
		end
	elseif grounded
		local vspeed = FixedMul(-speed, sin(ANG30))
		hand.momz = flip*vspeed + mo.momz
		hand.momx = mo.momx
		hand.momy = mo.momy
	else
		speed = $*3/4
		local vspeed = FixedMul(speed, sin(ANG30))
		local hspeed = FixedMul(speed, cos(ANG30))
		hand.momz = flip*vspeed + mo.momz
		hand.momx = mo.momx
		hand.momy = mo.momy
		P_Thrust(hand, hand.angle, hspeed)
	end
	
	if player.pflags&PF_SPINNING
		player.pflags = $&~PF_SPINNING
		mo.state = S_PLAY_FALL
	elseif player.pflags&PF_JUMPED and not (player.pflags&PF_NOJUMPDAMAGE)
		player.pflags = $|PF_NOJUMPDAMAGE
		mo.state = S_PLAY_FALL
	elseif mo.state == S_PLAY_NIGHTS_FLOAT
	or mo.state == S_TANGLE_FALLROLL
		mo.state = S_PLAY_FALL
	end
	if not grounded
		player.pflags = $|PF_THOKKED
		tangle.flags = $|TF_AIRBORNE
	end
	S_StartSound(hand, sfx_s3k82)
	
	if not grounded and not input
		hand.flags2 = $ | MF2_JUSTATTACKED // allow it to bust FOFs!
	end
	hand.prevmomentum = {x = hand.momx, y = hand.momy, z = hand.momz}
	
	tangle.hand = hand
end

// flag the hand as returning
local function GrappleLetGo(hand, noclip)
	if not S_SoundPlaying(hand, sfx_s3k5d)
	and not (valid(hand.tracer) and S_SoundPlaying(hand.tracer, sfx_s3k5d))
		S_StartSound(hand, sfx_prloop)
	end
	
	if valid(hand.tracer) and not (hand.flags2 & MF2_DONTRESPAWN) and not noclip
		hand.momx = $/2 + hand.tracer.momx/2
		hand.momy = $/2 + hand.tracer.momy/2
		hand.momz = $/2 + hand.tracer.momz/2
	end
	
	hand.flags = $ | MF_NOGRAVITY
	hand.flags2 = $ | MF2_BEYONDTHEGRAVE & ~MF2_DONTRESPAWN
	hand.frame = $ & ~FF_FRAMEMASK
	hand.target = nil
	hand.detectedside = nil
	hand.detectedline = nil
	hand.collidedline = nil
	hand.collidedside = nil
	hand.sidexscroll = nil
	hand.sideyscroll = nil
	hand.dumbfoftable = nil
	
	if noclip
		hand.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOCLIPTHING
	else
		hand.lastlook = 16
	end
end

// flag the hand as stuck to a surface
local function GrappleStick(hand, statechanges)
	//Why are we latching if we're already latched?
	if hand.flags2 & MF2_DONTRESPAWN
		return false
	end
	
	//Effects
	if not (hand.flags2 & MF2_BEYONDTHEGRAVE)
		for dust = 1, 4
			local dustjump = P_SpawnMobjFromMobj(hand, 0,0,0, MT_MINECARTSPARK)
			dustjump.fuse = TICRATE*2/3
			dustjump.scale = $/4
			P_InstaThrust(dustjump, (dust*ANGLE_90) + ANGLE_45 + P_RandomRange(1, FRACUNIT), 8*hand.scale)
			if statechanges == 2
				dustjump.momz = dustjump.momx
				P_InstaThrust(dustjump, hand.angle+ANGLE_90, dustjump.momy)
			end
		end
	end
	S_StopSoundByID(hand, sfx_s3k82)
	S_StopSoundByID(hand, sfx_prloop)

	//Break an fof if we can
	if hand.flags2 & MF2_JUSTATTACKED and BustGround(hand)
		hand.momx = hand.prevmomentum.x
		hand.momy = hand.prevmomentum.y
		hand.momz = hand.prevmomentum.z
		return false
	//No latching while homing, and no latching onto no climb walls
	elseif hand.tanglenoclimb or hand.flags2 & MF2_AMBUSH
	and not (valid(hand.target) and MF2_DONTRESPAWN)
		if not (hand.flags2 & MF2_BEYONDTHEGRAVE)
			GrappleLetGo(hand)
			S_StartSound(hand.tracer or hand, sfx_s3k5d)
		end
		return false
	end
	
	hand.flags = $ & ~MF_NOCLIP & ~MF_NOCLIPHEIGHT & ~MF_NOCLIPTHING
	hand.flags2 = $ | MF2_DONTRESPAWN &~ MF2_BEYONDTHEGRAVE
	hand.bouncesound = false
	hand.reactiontime = 5
	S_StartSound(hand, sfx_kc55)
	hand.momx = 0
	hand.momy = 0
	hand.momz = 0
	
	if statechanges
		//hand.state = hand.info.spawnstate
		for i = 1, statechanges
			hand.state = $//hand.flags2&MF2_OBJECTFLIP and hand.info.meleestate or hand.info.seestate
		end
	end
	
	return true
end

addHook("SpinSpecial", function(player)
	StartGrapple(player, PF_SPINDOWN)
end)

addHook("AbilitySpecial", function(player)
	StartGrapple(player, PF_JUMPDOWN)
end)

//Damage hook for the tail in BattleMod
addHook("ShouldDamage", function(mo,heart,owner)
	//Relegate this hook to interactions with players and heart projectiles
	if not CBW_Battle then return end
	if not(mo and mo.valid) then return end
	if not(mo.player) then return end
	if not(heart and heart.valid and heart.type == MT_TANGRAPPLE_SEGMENT) then return end
	if not(heart.flags&MF_MISSILE) then return end //heart is dead
	if not(owner and owner.type == MT_PLAYER and owner.player
	and not P_PlayerInPain(owner.player)) then return end
	//Player vs tagged
	if mo.player.pflags&PF_TAGIT and not(owner.player.pflags&PF_TAGIT) then return false end
	//Players
	if CBW_Battle.PlayerCanBeDamaged(mo.player) and not(CBW_Battle.MyTeam(owner.player,mo.player))
		then
		if G_TagGametype()
			return true
		elseif mo.player.battle_def or mo.player.battle_sdef
			if not mo.player.battle_def
				CBW_Battle.DoSPriority(mo.player, heart)
			end
			if mo.player.battle_def
				S_StartSound(mo,sfx_s3k7b)
				P_InstaThrust(mo,R_PointToAngle2(heart.x,heart.y,mo.x,mo.y),mo.scale*10)
				if heart.tracer and heart.tracer.valid
				and (not (heart.tracer.flags2 & (MF2_DONTRESPAWN|MF2_BEYONDTHEGRAVE))
				or heart.tracer.target == mo)
					GrappleLetGo(heart.tracer, true)
				end
				return false
			else
				return true
			end
		else
			return true
		end
	end
	return nil
end, MT_PLAYER)

//Detect springs/pushables/players for latching, and spikes for destroying
addHook("MobjMoveCollide", function(hand, mobj)
	if valid(hand) and valid(mobj) and ((hand.flags2 & MF2_AMBUSH
	and (mobj.flags&(MF_SPRING|MF_SOLID|MF_PUSHABLE)
	or (mobj.player and tanglepickup.value)) and not mobj.battleobject
	and mobj != hand.tracer and not (mobj.flags&MF_MONITOR)
	and not (hand.flags&MF_NOCLIPTHING) and not (hand.flags2 & MF2_DONTRESPAWN)
	and not (mobj.flags&MF_NOCLIPTHING) and mobj.health)
	or mobj.type == MT_SPIKE or mobj.type == MT_WALLSPIKE)
	and hand.z - hand.scale <= mobj.z + mobj.height
	and hand.z + hand.height + hand.scale >= mobj.z
		if mobj.type == MT_SPIKE or mobj.type == MT_WALLSPIKE
			P_KillMobj(mobj, hand, hand.tracer)
		else
			hand.target = mobj
			hand.flags2 = ($ & ~MF2_OBJECTFLIP) | (mobj.flags2 & MF2_OBJECTFLIP)
			hand.eflags = ($ & ~MFE_VERTICALFLIP) | (mobj.eflags & MFE_VERTICALFLIP)
			if valid(hand.tracer) and hand.tracer.player and hand.tracer.player.tangrapple
			and (hand.eflags & MFE_VERTICALFLIP) == (hand.tracer.eflags & MFE_VERTICALFLIP)
				hand.tracer.player.tangrapple.flags = $ | TF_FLOOR
			end
			GrappleStick(hand, 1)
		end
	end
end, MT_TANGRAPPLE_HAND)

//Detect spikes for destroying again
addHook("MobjMoveCollide", function(seg, mobj)
	if valid(seg) and valid(mobj) and valid(seg.target) and seg.target.player
	and seg.flags&MF_MISSILE and not (mobj.flags&MF_MISSILE)
	and (mobj.type == MT_SPIKE or mobj.type == MT_WALLSPIKE
	or mobj.type == MT_RING or mobj.type == MT_COIN
	or ((mobj.type == MT_FLINGRING or mobj.type == MT_FLINGCOIN) and mobj.fuse < 7*TICRATE)
	or (mobj.type == MT_REDTEAMRING and seg.target.player.ctfteam == 1)
	or (mobj.type == MT_BLUETEAMRING and seg.target.player.ctfteam == 2)
	or mobj.type == MT_NIGHTSSTAR or (CBW_Battle and mobj.battleobject))
	and seg.z - seg.scale <= mobj.z + mobj.height
	and seg.z + seg.height + seg.scale >= mobj.z
		if CBW_Battle and mobj.battleobject
			seg.pain_tics = 0
			local tracer = seg.tracer
			local isit = false
			if not (tracer and tracer.valid) return end
			if not tracer.bashlies then tracer.bashlies = {} end
			for i, bashee in ipairs(tracer.bashlies)
				if bashee == mobj then isit = true end
			end
			if not isit
				seg.momx = tracer.momx
				seg.momy = tracer.momy
				seg.momz = tracer.momz
				CBW_Battle.BashableCollision(mobj, seg)
				table.insert(tracer.bashlies, mobj)
			elseif seg and seg.valid
				seg.pain_tics = TICRATE
			end
		else
			if mobj.type != MT_SPIKE and mobj.type != MT_WALLSPIKE
				P_GivePlayerRings(seg.target.player, 1)
				if mobj.info.deathsound == sfx_rnglft
				or mobj.info.deathsound == sfx_rngrgh
					mobj.info.deathsound = sfx_itemup
				end
				S_StartSound(mobj, mobj.info.deathsound)
				mobj.momx = 0
				mobj.momy = 0
				mobj.momz = 0
				
				if mapheaderinfo[gamemap].typeoflevel&TOL_NIGHTS
				and mobj.type != MT_FLINGRING and mobj.type != MT_FLINGCOIN
					P_DoNightsScore(seg.target.player)
				end
			end
			P_KillMobj(mobj, seg, seg.target)
		end
	end
end, MT_TANGRAPPLE_SEGMENT)

//Detect walls for latching
addHook("MobjMoveBlocked", function(hand)
	if not hand.valid
	or hand.flags2 & MF2_DONTRESPAWN
	or (hand.flags2 & MF2_BEYONDTHEGRAVE
	and not (valid(hand.tracer) and hand.tracer.player and hand.tracer.player.tangrapple
	and (hand.tracer.player.pflags & PF_SPINDOWN or hand.tracer.player.pflags & PF_JUMPDOWN
	and hand.tracer.player.tangrapple.flags & TF_COUNTJUMP)))
		return
	end
	
	if GrappleStick(hand, 2) and not valid(hand.target)
	and valid(hand.detectedside) and valid(hand.detectedline)
		local detectedline = hand.detectedline
		local detectedside = hand.detectedside
		for i=1, 8
			P_TryMove(hand, hand.x+hand.prevmomentum.x/16, hand.y+hand.prevmomentum.y/16, true)
		end
		
		hand.collidedline = detectedline
		hand.collidedside = detectedside
		hand.sidexscroll = hand.collidedside.textureoffset
		hand.sideyscroll = hand.collidedside.rowoffset
		hand.dumbfoftable = {}
		if valid(hand.collidedline.frontsector)
			table.insert(hand.dumbfoftable, {hand.collidedline.frontsector, hand.collidedline.frontsector.ceilingheight, hand.collidedline.frontsector.floorheight})
			for fof in hand.collidedline.frontsector.ffloors()
				if valid(fof)
					table.insert(hand.dumbfoftable, {fof, fof.topheight, fof.bottomheight})
				end
			end
		end
		if valid(hand.collidedline.backsector)
			table.insert(hand.dumbfoftable, {hand.collidedline.backsector, hand.collidedline.backsector.ceilingheight, hand.collidedline.backsector.floorheight})
			for fof in hand.collidedline.backsector.ffloors()
				if valid(fof)
					table.insert(hand.dumbfoftable, {fof, fof.topheight, fof.bottomheight})
				end
			end
		end
	end
end, MT_TANGRAPPLE_HAND)

//Detect walls
addHook("MobjLineCollide", function(hand, line)
	if valid(line) and valid(hand)
		if hand.flags&MF_NOCLIP and hand.flags&MF_NOCLIPHEIGHT
			return false
		elseif fofdetect(hand, line.frontsector, FF_BUSTUP)
		or fofdetect(hand, line.backsector, FF_BUSTUP)
			hand.flags2 = $ & ~MF2_DONTRESPAWN
			hand.frame = $ & ~FF_FRAMEMASK
			hand.tanglenoclimb = true
			return false
		end
		
		hand.detectedline = line
		hand.detectedside = sides[line.sidenum[P_PointOnLineSide(hand.x, hand.y, line)]]
		
		if line.flags & ML_NOCLIMB or line.special == 41
		or (valid(line.frontsector) and line.frontsector.ceilingpic == "F_SKY1"
		and line.frontsector.floorheight == line.frontsector.ceilingheight
		and hand.z > line.frontsector.ceilingheight)
		or (valid(line.backsector) and line.backsector.ceilingpic == "F_SKY1"
		and line.backsector.floorheight == line.backsector.ceilingheight
		and hand.z > line.backsector.ceilingheight)
			hand.detectedline = nil
			hand.detectedside = nil
			hand.tanglenoclimb = true
		else
			hand.tanglenoclimb = false
		end
		
		return fofdetect(hand, line.frontsector, FF_BLOCKPLAYER)
		or fofdetect(hand, line.backsector, FF_BLOCKPLAYER)
	end
end, MT_TANGRAPPLE_HAND)

//Prevent going into the ERZ1 spin path as Tangle
addHook("MobjLineCollide", function(mo, line)
	if gamemap == 22 and #line >= 211 and #line <= 214
	and mapheaderinfo[gamemap].lvlttl == "Egg Rock"
	and mapheaderinfo[gamemap].actnum == 1
	and mo and mo.valid and mo.skin == SKIN and mo.player
		if mo.player.pflags&PF_SPINNING
			mo.momx = 0
			mo.momy = 0
			mo.player.pflags = $&~PF_SPINNING
			mo.state = S_PLAY_WALK
		end
		return true
	end
end, MT_PLAYER)

addHook("PlayerQuit", function(player)
	local tangle = player.tangrapple
	if tangle
		player.tangrapple = nil
		if valid(tangle.hand) and tangle.hand.type == MT_TANGRAPPLE_HAND
			P_RemoveMobj(tangle.hand)
		end
		if valid(tangle.tail) and tangle.tail.type == MT_TANGRAPPLE_TAIL
			P_RemoveMobj(tangle.tail)
		end
	end
	if player.tanglesegs
		for _, seg in ipairs(player.tanglesegs)
			if valid(seg)
				P_RemoveMobj(seg)
			end
		end
		player.tanglesegs = nil
	end
end)

addHook("ThinkFrame", do for player in players.iterate
	local mo = player.mo
	local tangle = player.tangrapple
	
	if not mo
	or mo.skin ~= SKIN
	or player.quittime
		if tangle
			player.tangrapple = nil
			if valid(tangle.hand) and tangle.hand.type == MT_TANGRAPPLE_HAND
				P_RemoveMobj(tangle.hand)
			end
			if valid(tangle.tail) and tangle.tail.type == MT_TANGRAPPLE_TAIL
				P_RemoveMobj(tangle.tail)
			end
		end
		if player.tanglesegs
			for _, seg in ipairs(player.tanglesegs)
				if valid(seg)
					P_RemoveMobj(seg)
				end
			end
			player.tanglesegs = nil
		end
		continue
	end
	
	//Apparently Tangle & Whisper sets Tangle's height incorrectly!
	//Thanks SRB2
	if not multiplayer and #players > 1 and leveltime == 1
		player.height = skins[player.skin].height
	end
	
	//Yes we're doing this here, don't ask
	if S_SoundPlaying(mo, sfx_jump) and mo.triplethreshold
		if mo.triplethreshold >= 12
			S_StartSound(mo, sfx_tngjm3)
			S_StopSoundByID(mo, sfx_jump)
		elseif mo.triplethreshold >= 8
			S_StartSound(mo, sfx_tngjm2)
			S_StopSoundByID(mo, sfx_jump)
		end
	end
	
	//Make you face the camera with death sprites so as to make the tail look correct
	if mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_DEAD
	or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_DRWN
		player.drawangle = (player.cmd.angleturn<<16)-ANGLE_180
	end
	
	// variable setup
	if not tangle
	or mo ~= tangle.mo
		tangle = {}
		player.tangrapple = tangle
		tangle.player = player
		tangle.mo = mo
		
		tangle.flags = 0
		tangle.prevPosition = {x = mo.x, y = mo.y, z = mo.z}
	end
	
	local hand = tangle.hand
	local tail = tangle.tail
	local grounded = P_IsObjectOnGround(mo)
	local oldspeed = FixedHypot(FixedHypot(mo.momx, mo.momy), abs(mo.momz))
	
	//Spawn lockon
	local lockon = EnemyLookDeluxe(player, THROW_DISTANCE*5/4, ANGLE_22h)
	if lockon and lockon.valid and PlayerCanGrapple(player) and not valid(hand)
	and not PlayerFullInput(player) and (grounded or player.pflags&PF_JUMPED)
		P_SpawnLockOn(player, lockon, S_LOCKON3)
	end
	
	// unmark airborne flag on ground touch
	if grounded and not tangle.grounded
	or mo.eflags & MFE_SPRUNG
		tangle.flags = $ & ~TF_AIRBORNE
	end
	
	local forwardoffset = FixedDiv(-5*mo.radius/2, mo.scale)
	local verticaloffset = FixedDiv(mo.height*2/6, mo.scale)
	local thrustx = cos(player.drawangle)
	local thrusty = sin(player.drawangle)
	local threshold = FixedMul(STRETCH_THRESHOLD, mo.scale)
	
	if valid(tail) //The tail object
		tail.fuse = 3 //Delete it if we're not setting this constantly
		if valid(hand)
			tail.state = S_INVISIBLE
			if player.midtanglepoint
				P_TeleportMove(tail, player.midtanglepoint.x, player.midtanglepoint.y, player.midtanglepoint.z)
			else
				P_TeleportMove(tail, hand.x, hand.y, hand.z)
			end
			tail.momx = hand.momx
			tail.momy = hand.momy
			tail.momz = hand.momz
		elseif mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_ROLL
			tail.state = S_INVISIBLE
			P_TeleportMove(tail, mo.x, mo.y, mo.z)
			tail.momx = mo.momx
			tail.momy = mo.momy
			tail.momz = mo.momz
		elseif mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_TRNS
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT1
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT2
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT3
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NDRL
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NFLY
		or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NPUL
		or FixedHypot(FixedHypot(tail.x - mo.x, tail.y - mo.y), (tail.z + (tail.height/2)) - (mo.z + (mo.height/2))) > threshold*2
			tail.state = S_INVISIBLE
			P_TeleportMove(tail, mo.x + FixedMul(FixedMul(forwardoffset, thrustx), mo.scale),
			mo.y + FixedMul(FixedMul(forwardoffset, thrusty), mo.scale),
			mo.z + FixedMul(verticaloffset, mo.scale)*P_MobjFlip(mo))
			tail.momx = mo.momx
			tail.momy = mo.momy
			tail.momz = mo.momz
		else
			local speed = max(FixedDiv(player.speed, mo.scale)*5/4, 5*FRACUNIT)
			if player.powers[pw_carry]&~CR_NIGHTSMODE&~CR_NIGHTSFALL
				speed = FRACUNIT/16
			end

			if player.powers[pw_carry] == CR_MINECART
			or player.powers[pw_carry] == CR_ROLLOUT
			or (grounded and not (player.powers[pw_carry]&~CR_NIGHTSMODE&~CR_NIGHTSFALL))
				tail.reactiontime = $+speed
			elseif tail.reactiontime > 90*FRACUNIT and tail.reactiontime < 270*FRACUNIT
				tail.reactiontime = (($-180*FRACUNIT)*3/4)+180*FRACUNIT
			elseif tail.reactiontime == 180*FRACUNIT
				tail.reactiontime = 0
			else
				tail.reactiontime = $*3/4
			end
			
			if tail.reactiontime >= 360*FRACUNIT
				tail.reactiontime = $-360*FRACUNIT
			end
			
			if tail.threshold != -4*mo.momz*P_MobjFlip(mo)
				tail.threshold = min(max(($+4*mo.momz*P_MobjFlip(mo))*2/3 - 4*mo.momz*P_MobjFlip(mo), -32*mo.scale), 32*mo.scale)
			end
			
			verticaloffset = $+FixedDiv(16 * sin(FixedAngle(tail.reactiontime)), max(speed/16,FRACUNIT))+FixedDiv(tail.threshold, mo.scale)
			forwardoffset = $+abs(FixedSquare(verticaloffset)-FixedSquare(FixedDiv(mo.height*2/5, mo.scale)))/128
			
			if not (tail.verticaloffsethistory and #tail.verticaloffsethistory)
				tail.verticaloffsethistory = {}
			end
			table.insert(tail.verticaloffsethistory, verticaloffset)
			if #tail.verticaloffsethistory > (grounded and max(7-(player.speed/80/FRACUNIT), 1) or 1)
				table.remove(tail.verticaloffsethistory, 1)
			end
		
			local x = mo.x + FixedMul(forwardoffset, FixedMul(thrustx, mo.scale))
			local y = mo.y + FixedMul(forwardoffset, FixedMul(thrusty, mo.scale))
			local z = mo.z + FixedMul(tail.verticaloffsethistory[1], mo.scale)*P_MobjFlip(mo)
			
			local hdistance = FixedHypot(tail.x - x, tail.y - y)
			local vdistance =  (tail.z + (tail.height/2)) - (z + (mo.height/2))
			local distance = FixedHypot(hdistance, vdistance)
			
			local cap = FixedMul(STRETCH_CAP/16, mo.scale)
			local factor = FixedDiv(distance - threshold/2, cap - threshold/2)
			factor = max(min($, FRACUNIT), FRACUNIT/8)
			
			tail.momx = $/2 + FixedMul(x - tail.x, factor) + mo.momx*2/5
			tail.momy = $/2 + FixedMul(y - tail.y, factor) + mo.momy*2/5
			if tail.eflags & MFE_VERTICALFLIP
				tail.momz = $/2 - FixedMul((tail.z + tail.height) - (z + mo.height), factor) + mo.momz*2/5
			else
				tail.momz = $/2 - FixedMul(tail.z - z, factor) + mo.momz*2/5
			end
			
			if tail.state != tail.info.seestate
				tail.state = tail.info.seestate
			end
			tail.scale = mo.scale/2
			tail.color = mo.color
			tail.flags2 = ($&~MF2_DONTDRAW)|(mo.flags2&MF2_DONTDRAW)
			tail.flags2 = ($&~MF2_OBJECTFLIP)|(mo.flags2&MF2_OBJECTFLIP)
			tail.eflags = ($&~MFE_VERTICALFLIP)|(mo.eflags&MFE_VERTICALFLIP)
			tail.frame = ($&~FF_FULLBRIGHT)|(mo.frame&FF_FULLBRIGHT)
			
			RenderTail(player, tail, grounded and FixedMul((FixedSquare(verticaloffset)-FixedSquare(FixedDiv(mo.height*2/5, mo.scale)))*16*P_MobjFlip(mo),mo.scale) or 0)
		end
	else
		tangle.tail = P_SpawnMobjFromMobj(mo,
		FixedMul(forwardoffset, thrustx), FixedMul(forwardoffset, thrusty),
		verticaloffset, MT_TANGRAPPLE_TAIL)
		tail = tangle.tail
		tail.skin = TAILSKIN
	end
	
	if valid(hand) // the player activated their grapple and spawned a tail!
		hand.fuse = 3 //Delete it if we're not setting this constantly
		
		//Special stand animation
		if (player.pflags & PF_SPINDOWN or player.pflags & PF_JUMPDOWN and tangle.flags & TF_COUNTJUMP)
		and grounded and player.panim == PA_IDLE
			mo.state = S_PLAY_GLIDE_LANDING
			mo.tics = 1
			player.panim = PA_IDLE
		end
		
		//Decrease latch timer
		if hand.reactiontime
			hand.reactiontime = $-1
		end
		if hand.threshold
			hand.threshold = $-1
		end
		if hand.lastlook
			hand.lastlook = $-1
			if not hand.lastlook
			and hand.flags2 & MF2_BEYONDTHEGRAVE and not (hand.flags&MF_NOCLIP)
				hand.flags = $|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
			end
		end
		
		//Go against goop gravity
		if P_IsObjectInGoop(hand) and not (hand.flags&MF_NOGRAVITY)
			hand.momz = $-P_GetMobjGravity(hand)
			hand.momz = $-FixedMul(gravity*P_MobjFlip(hand), hand.scale)
		end
	
		//Go to the object we're latched to if it exists, and pull it if it's pushable
		if hand.target and hand.flags2 & MF2_DONTRESPAWN
			if not hand.target.valid or not hand.target.health
			or (not hand.reactiontime and hand.target.player and not (gametyperules&GTR_FRIENDLY)
			and (not G_GametypeHasTeams() or hand.target.ctfteam != hand.target.player.ctfteam)
			and not P_PlayerInPain(hand.target.player))
				GrappleLetGo(hand, true)
			else
				P_TeleportMove(hand, hand.target.x, hand.target.y, hand.target.z+hand.target.height/2)
				
				if hand.target.type == MT_ROLLOUTROCK
					hand.target.fuse = hand.target.info.painchance
				end
				
				if hand.target.battleobject or hand.target.flags&MF_MISSILE
					hand.target.target = mo
				end
				
				local hdistance = FixedHypot(hand.target.x - mo.x, hand.target.y - mo.y)
				local vdistance = (hand.target.z + (hand.target.height >> 1)) - (mo.z + (mo.height >> 1))
				local distance = FixedHypot(hdistance, vdistance)
				
				if (hand.target.flags&MF_PUSHABLE or hand.target.player)
				and distance > threshold
					local cap = FixedMul(STRETCH_CAP, mo.scale)
					local factor = FixedDiv(distance - threshold, cap - threshold)
					if hand.target.type != MT_ROLLOUTROCK
						if hand.target.player
							if P_GetPlayerControlDirection(hand.target.player)
							or not (grounded and P_IsObjectOnGround(hand.target))
								factor = FixedSquare($)
							end
						else
							factor = FixedSquare($)*2
						end
					end
					if factor != 0 and hand.target.player
					and P_GetPlayerControlDirection(player)
					and (player.powers[pw_super]
					or player.powers[pw_sneakers])
					and not (hand.target.player.powers[pw_super]
					or hand.target.player.powers[pw_sneakers])
						factor = FixedDiv($,$)
					end
					factor = max(min($, FRACUNIT/64), 0)
					
					hand.target.momx = $ + FixedMul(mo.x - hand.target.x, factor)
					hand.target.momy = $ + FixedMul(mo.y - hand.target.y, factor)
					if not P_IsObjectOnGround(hand.target) or abs(vdistance) > threshold*3
						if hand.target.eflags & MFE_VERTICALFLIP
							hand.target.momz = $ + FixedMul((mo.z + mo.height) - (hand.target.z + hand.target.height), factor)
						else
							hand.target.momz = $ + FixedMul(mo.z - hand.target.z, factor)
						end
						if P_IsObjectOnGround(hand.target) and hand.target.player
							if hand.target.player.panim ~= PA_ETC
							and hand.target.player.panim <= PA_DASH
								hand.target.state = S_PLAY_FALL
							end
						end
					end
				end
			end
		end
		
		local hdistance = FixedHypot(mo.x - hand.x, mo.y - hand.y)
		local vdistance = (mo.z + (mo.height >> 1)) - (hand.z + (hand.height >> 1))
		local distance = FixedHypot(hdistance, vdistance)
		
		// if the player releases their buttons or the hand is too far away, retract the tail
		if not (hand.flags2 & MF2_BEYONDTHEGRAVE)
			if not (player.pflags & PF_SPINDOWN or player.pflags & PF_JUMPDOWN and tangle.flags & TF_COUNTJUMP)
			or (distance > FixedMul(hand.flags2&MF2_AMBUSH and THROW_DISTANCE*5/4 or THROW_DISTANCE, mo.scale)
			and not hand.reactiontime and not (hand.flags2&MF2_DONTRESPAWN) and not player.powers[pw_super])
			or distance > FixedMul(THROW_DISTANCE*2, mo.scale) or not PlayerCanGrapple(player)
				
				//Extra bounce and spinning from releasing manually
				if not (player.pflags & PF_SPINDOWN or player.pflags & PF_JUMPDOWN and tangle.flags & TF_COUNTJUMP)
				and hand.flags2 & MF2_DONTRESPAWN and PlayerCanGrapple(player)
				and not (hand.flags & MF_NOCLIP) and distance <= FixedMul(THROW_DISTANCE*2, mo.scale)
					GrappleLetGo(hand, true)
					mo.state = S_TANGLE_FALLROLL
					player.panim = PA_FALL
					
					if not grounded or hand.threshold
					//or distance <= FixedMul(THROW_DISTANCE, mo.scale)
						local momzlimit = 24*mo.scale
						local newmomz = vdistance/-max(abs(FixedHypot(player.speed, mo.momz)/mo.scale/2), 10)
						if hand.threshold
							newmomz = $*2
						end
						newmomz = FixedDiv($, FixedDiv(max(hdistance/128,mo.scale), mo.scale))
						
						if grounded and hdistance > 8*mo.radius
						and abs(vdistance) < (tangle.flags&TF_FLOOR and mo.height*2 or mo.height)
							player.pflags = $|PF_SPINNING
							mo.state = S_PLAY_ROLL
							S_StartSound(mo, sfx_zoom)
							P_InstaThrust(mo, R_PointToAngle2(mo.x, mo.y, hand.x, hand.y), min(hdistance/(player.gotflagdebuff and 10 or 6), 75*mo.scale))
						elseif not grounded or abs(newmomz) > mo.height/4
							player.pflags = $|P_GetJumpFlags(player)|PF_THOKKED
							
							if abs(mo.momz) <= momzlimit
							and abs(mo.momz) > abs(newmomz/2)
								mo.momz = min(max($+newmomz, -momzlimit), momzlimit)
							elseif abs(mo.momz) <= abs(newmomz/2)
							and abs(mo.momz) <= momzlimit
								mo.momz = min(max(newmomz/2, -momzlimit), momzlimit)
							end
							
							if not player.gotflagdebuff
								mo.momx = $*9/8
								mo.momy = $*9/8
							end
						end
					end
				elseif not (hand.flags2 & MF2_DONTRESPAWN)
					GrappleLetGo(hand)
					hand.z = $+hand.momz
					P_TryMove(hand, hand.x+hand.momx, hand.y+hand.momy, true)
				else
					GrappleLetGo(hand)
				end
			end
		end
		
		// if the player releases jump, don't count jump as a relevant grapple button
		if not (player.pflags & PF_JUMPDOWN)
		and tangle.flags & TF_COUNTJUMP
			tangle.flags = $ & ~TF_COUNTJUMP
		end
		
		// grapple thinker
		if hand.flags2 & MF2_BEYONDTHEGRAVE // retracting, gotta return to the player
			if distance < threshold
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_ROLL
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_TRNS
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT1
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT2
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_CNT3
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NDRL
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NFLY
			or mo.sprite2&~FF_SPR2SUPER&~FF_SPR2ENDSTATE&~FF_SPR2MIDSTART == SPR2_NPUL
			or distance > FixedMul(THROW_DISTANCE*3, mo.scale)
				P_RemoveMobj(hand)
				mo.tailjectile = false // Bloops' variable for hiding the tail
				tangle.hand = nil
				continue
			else
				local angle = R_PointToAngle2(hand.x, hand.y, mo.x, mo.y)
				local aiming = R_PointToAngle2(0, hand.z + (hand.height >> 1), hdistance, mo.z + (mo.height >> 1))
				local speed = max(FixedMul(hand.info.speed, mo.scale), FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz) + 16*mo.scale)
				local hthrust = FixedMul(speed, cos(aiming))
				local divfactor = (player.powers[pw_super] and 2 or 4)
				hand.momx = $ + (P_ReturnThrustX(mo, angle, hthrust) - $) / divfactor
				hand.momy = $ + (P_ReturnThrustY(mo, angle, hthrust) - $) / divfactor
				hand.momz = $ + (FixedMul(speed, sin(aiming)) - $) / divfactor
				hand.angle = angle + ANGLE_180
			end
		end
		
		if PlayerCanGrapple(player) and not (hand.eflags & MFE_TOUCHLAVA
		and not (player.powers[pw_shield] & SH_PROTECTFIRE)
		and not player.powers[pw_invulnerability] and not player.powers[pw_super]
		and not player.powers[pw_flashing] and not (player.pflags & PF_GODMODE))
			if hand.flags2 & MF2_DONTRESPAWN // grappling, fuck with the player's momentum here
				if distance > threshold and distance <= FixedMul(THROW_DISTANCE, mo.scale)*2
					local cap = FixedMul(STRETCH_CAP, mo.scale)
					local factor = FixedDiv(distance - threshold, cap - threshold)
					
					if player.cmd.forwardmove or player.cmd.sidemove
					or not (grounded and (tangle.flags & TF_FLOOR
					or (valid(hand.target) and P_IsObjectOnGround(hand.target))))
						factor = FixedSquare($)
					end
					if factor != 0 and valid(hand.target) and hand.target.player
					and P_GetPlayerControlDirection(hand.target.player)
					and (hand.target.player.powers[pw_super]
					or hand.target.player.powers[pw_sneakers])
					and not (player.powers[pw_super]
					or player.powers[pw_sneakers])
						factor = FixedDiv($,$)
					end
					factor = max(min($, FRACUNIT/64), 0)
					
					mo.momx = $ + FixedMul(hand.x - mo.x, factor)
					mo.momy = $ + FixedMul(hand.y - mo.y, factor)
					if not grounded or abs(vdistance) > threshold*3
						if factor != 0 and valid(hand.target)
						and hand.target.player and hand.target.player.pflags & PF_GLIDING
							factor = $*2
						end
						if mo.eflags & MFE_VERTICALFLIP
							mo.momz = $ + FixedMul((hand.z + hand.height) - (mo.z + mo.height), factor)
						else
							mo.momz = $ + FixedMul(hand.z - mo.z, factor)
						end
						if (tangle.grounded or grounded) and not (tangle.flags & TF_AIRBORNE)
							tangle.flags = $ | TF_AIRBORNE
							if player.panim ~= PA_ETC and player.panim <= PA_DASH
							and not (player.pflags & (PF_JUMPED|PF_SPINNING))
								mo.state = S_PLAY_FALL
							end
						end
					end
					
					local newspeed = FixedHypot(FixedHypot(mo.momx, mo.momy), abs(mo.momz))
					if FixedHypot(oldspeed, newspeed) > 16*mo.scale
					and ((oldspeed > newspeed*16/15 and oldspeed > newspeed)
					or (newspeed > oldspeed*128/127 and newspeed < oldspeed))
						if not hand.bouncesound
							S_StartSoundAtVolume(mo, sfx_kc52, 72)
							hand.threshold = TICRATE/2
						end
						hand.bouncesound = true
					else
						hand.bouncesound = false
					end
				end
				
				//Move with conveyors and fofs
				if valid(hand.collidedside) and valid(hand.collidedline)
				and hand.sidexscroll != nil and hand.sideyscroll != nil
					//Move with conveyors
					if hand.collidedside.textureoffset != hand.sidexscroll
					or hand.collidedside.rowoffset != hand.sideyscroll
						local angle = R_PointToAngle2(hand.collidedline.v1.x, hand.collidedline.v1.y,
						hand.collidedline.v2.x, hand.collidedline.v2.y)
						if hand.collidedline.frontside == hand.collidedside
							angle = $-ANGLE_180
						end
						
						//Horizontal movement
						P_TryMove(hand,
						hand.x+FixedMul(hand.collidedside.textureoffset-hand.sidexscroll, cos(angle)),
						hand.y+FixedMul(hand.collidedside.textureoffset-hand.sidexscroll, sin(angle)),
						true)
						
						//Verical Movement
						hand.z = $+hand.collidedside.rowoffset-hand.sideyscroll
						
						//Update variables
						hand.sidexscroll = hand.collidedside.textureoffset
						hand.sideyscroll = hand.collidedside.rowoffset
						
						//Make sure we're still on the wall
						local pointx,pointy = P_ClosestPointOnLine(hand.x, hand.y, hand.collidedline)
						angle = R_PointToAngle2(hand.x, hand.y, pointx, pointy)
						
						if P_TryMove(hand, hand.x+FixedMul(hand.radius*6,cos(angle)),
						hand.y+FixedMul(hand.radius*6,sin(angle)), true)
							GrappleLetGo(hand)
						end
					end
						
					//Move with fofs
					if hand.dumbfoftable and #hand.dumbfoftable
						for id, fof in pairs(hand.dumbfoftable)
							local sector = userdataType(fof[1]) == "sector_t"
							if valid(fof[1]) and (sector or fof[1].flags & FF_BLOCKPLAYER)
								local topheight = 0
								local bottomheight = 0
								local t_slope = 0
								local b_slope = 0
								//Yes this is actually for sectors as well
								if sector
									topheight = fof[1].ceilingheight
									bottomheight = fof[1].floorheight
									t_slope = fof[1].c_slope
									b_slope = fof[1].f_slope
								else
									topheight = fof[1].topheight
									bottomheight = fof[1].bottomheight
									t_slope = fof[1].t_slope
									b_slope = fof[1].b_slope
								end
								
								//Is there a difference in fof height?
								if topheight != fof[2] or bottomheight != fof[3]
									//Let's make sure we're actually touching it
									local foffloor = topheight
									local fofceiling = bottomheight
									if t_slope and t_slope.valid
										foffloor = P_GetZAt(t_slope, hand.x, hand.y)
									end
									if b_slope and b_slope.valid
										fofceiling = P_GetZAt(b_slope, hand.x, hand.y)
									end
									
									//Apply movement if so
									if foffloor >= hand.z and fofceiling <= hand.z+hand.height
										if topheight != fof[2] and bottomheight != fof[3]
											hand.z = $+((topheight-fof[2])/2)+((bottomheight-fof[3])/2)
										else
											hand.z = $+(topheight-fof[2])+(bottomheight-fof[3])
										end
									end
								end
								
								//Update variables
								fof[2] = topheight
								fof[3] = bottomheight
							end
						end
					end
				end
				
				if hand.flags2 & MF2_BEYONDTHEGRAVE
					hand.flags2 = $&~MF2_DONTRESPAWN
				end
				
			// extending, break floors/ceilings or attach to them
			elseif not (hand.flags & (MF_NOCLIP|MF_NOCLIPHEIGHT))
			and valid(hand.subsector) and valid(hand.subsector.sector)
			and not (hand.flags2 & MF2_AMBUSH)
			and (player.pflags & PF_SPINDOWN or player.pflags & PF_JUMPDOWN and tangle.flags & TF_COUNTJUMP)
				local fofcollide = fofdetect(hand, hand.subsector.sector, FF_BLOCKPLAYER)
				local onfloor = hand.z + hand.momz <= hand.floorz
				and (hand.subsector.sector.floorpic != "F_SKY1" or fofcollide)
				local onceiling = hand.z + hand.height + hand.momz >= hand.ceilingz
				and (hand.subsector.sector.ceilingpic != "F_SKY1" or fofcollide)
				
				if (onfloor or onceiling) and GrappleStick(hand, 1)
					if onfloor
						hand.flags2 = $ & ~MF2_OBJECTFLIP
						hand.eflags = $ & ~MFE_VERTICALFLIP
					else
						hand.flags2 = $ | MF2_OBJECTFLIP
						hand.eflags = $ | MFE_VERTICALFLIP
					end
					if (hand.eflags & MFE_VERTICALFLIP) == (mo.eflags & MFE_VERTICALFLIP)
						tangle.flags = $ | TF_FLOOR
					end
					hand.flags = $ & ~MF_NOGRAVITY
				end
				
				if (leveltime%2 or (hand.flags2 & MF2_DONTRESPAWN and (onfloor or onceiling)))
				and not (hand.flags2 & MF2_BEYONDTHEGRAVE)
					for dust = 1, 8
						local dustjump = P_SpawnMobjFromMobj(hand,
						FixedDiv(hand.momx, hand.scale), FixedDiv(hand.momy, hand.scale),
						FixedDiv(hand.momz*P_MobjFlip(hand), hand.scale), MT_DUST)
						dustjump.flags2 = $|MF2_SHADOW
						dustjump.state = states[$].nextstate
						P_InstaThrust(dustjump, hand.angle + (dust*ANGLE_45) + (leveltime%4)*ANGLE_22h, 6*hand.scale)
						if FixedHypot(hand.momx, hand.momy) > abs(hand.momz)
						and not (hand.flags2 & MF2_DONTRESPAWN and (onfloor or onceiling))
							dustjump.momz = dustjump.momx
							P_InstaThrust(dustjump, R_PointToAngle2(0,0,hand.momx, hand.momy)+ANGLE_90, dustjump.momy)
						end
					end
				end
			end
		elseif not (hand.flags2 & MF2_BEYONDTHEGRAVE)
			GrappleLetGo(hand, true)
		end
		
		hand.prevmomentum = {x = hand.momx, y = hand.momy, z = hand.momz}
		
		// visuals
		hand.scale = mo.scale
		hand.color = mo.color
		//hand.flags2 = ($&~MF2_DONTDRAW)|(mo.flags2&MF2_DONTDRAW)
		hand.frame = ($&~FF_FULLBRIGHT)|(mo.frame&FF_FULLBRIGHT)
		
		RenderTail(player, hand)
	end
	
	tangle.grounded = grounded
end end)

//Battle Mod Support
local ScriptLoaded = false
local creditsWait = TICRATE
addHook("ThinkFrame", function()
	//Credits
	if creditsWait
		creditsWait = $ - 1
		if not creditsWait
			print("\140Tangle the Lemur v1.0.2",
			"\130Sprites by Lightdasher, Chrispy\128 - \135Code by Bloops, Lach, and SMS Alfredo",
			"\132Extra help from Nibroc.rock and Frostiikin",
			"\134Try out the \128tangleswap \134and \128tanglepickup \134commands!")
		end
	end
	
	if not(CBW_Battle) or ScriptLoaded then return end
	ScriptLoaded = true
	local B = CBW_Battle
	
	//Stats
	B.SkinVars[SKIN] = {
		func_exhaust = function(player)
			if player.pflags&PF_THOKKED or (player.tangrapple and valid(player.tangrapple.hand)
			and not (player.tangrapple.hand.flags2&MF2_BEYONDTHEGRAVE))
				local maxtime = 6*TICRATE
				player.exhaustmeter = max(0,$-FRACUNIT/maxtime)
				if player.tangrapple and valid(player.tangrapple.hand)
					if not (player.tangrapple.hand.flags2&MF2_BEYONDTHEGRAVE)
						if player.exhaustmeter <= 0
							GrappleLetGo(player.tangrapple.hand, true)
						end
					elseif player.exhaustmeter <= 0
					and not (player.tangrapple.hand.flags&(MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING))
						player.tangrapple.hand.flags = $|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOCLIPTHING
					end
				end
			elseif P_IsObjectOnGround(player.mo)
				player.exhaustmeter = FRACUNIT
			end
			return true
		end
	}
end)